from tkinter import *
import random
import time

# Constants
SCREEN_WIDTH = 480
SCREEN_HEIGHT = 270
SHIP_SPEED = 6
ENEMY_BASE_SPEED = 2
ENEMY_BASE_SIZE = 20
SCORE_PER_WAVE = 100
WAVE_INCREMENT = 50
ENEMY_SPAWN_INCREMENT = 1  # Increment for spawning one more enemy each wave
ENEMY_SPEED_INCREMENT = 0.5  # Increment for increasing enemy speed


class Ship:
    def __init__(self):
        self.x = SCREEN_WIDTH // 2
        self.y = SCREEN_HEIGHT // 2
        self.size = 32
        self.image = PhotoImage(file="spaceship.gif")
        self.has_gun = False  # Initially the ship doesn't have a gun
        self.shots_remaining = 2  # Number of shots remaining before cooldown
        self.cooldown_timer = 0  # Cooldown timer for the gun
        self.last_shot_time = 0  # Time when the last shot was fired

    def shoot(self):
        if self.has_gun and self.shots_remaining > 0 and time.time() - self.last_shot_time >= 5:
            self.shots_remaining -= 1
            self.last_shot_time = time.time()
            for enemy in enemies:
                if (self.x < enemy.x + enemy.size and self.x + self.size > enemy.x
                    and self.y < enemy.y + enemy.size and self.y + self.size > enemy.y):
                    enemies.remove(enemy)
                    self.shots_remaining += 1  # Resetting shots if an enemy is hit

    def update_cooldown(self):
        if time.time() - self.last_shot_time >= 5:
            self.shots_remaining = 2


class Enemy:
    def __init__(self, size, speed):
        self.x = random.randint(0, SCREEN_WIDTH)
        self.y = random.randint(-100, -50)
        self.size = size
        self.speed = speed
        self.image = PhotoImage(file="alien.gif")

def update():
    global ship, total_score, wave_number, enemies, start_time, elapsed_time

    # Calculate elapsed time
    elapsed_time = time.time() - start_time

    # Move the player's ship based on key presses
    if keys['d']:
        ship.x += SHIP_SPEED
    if keys['a']:
        ship.x -= SHIP_SPEED
    if keys['s']:
        ship.y += SHIP_SPEED
    if keys['w']:
        ship.y -= SHIP_SPEED

    # Clamp ship position within the screen
    ship.x = max(0, min(SCREEN_WIDTH - ship.size, ship.x))
    ship.y = max(0, min(SCREEN_HEIGHT - ship.size, ship.y))

    # Update enemies
    for enemy in enemies:
        enemy.y += enemy.speed

        # Check for collisions with the player's ship
        if (ship.x < enemy.x + enemy.size and ship.x + ship.size > enemy.x
                and ship.y < enemy.y + enemy.size
                and ship.y + ship.size > enemy.y):
            game_over()
            return

        # Respawn enemies if they reach the bottom of the screen
        if enemy.y > SCREEN_HEIGHT:
            # Find a non-overlapping position for the new enemy
            while True:
                new_enemy_x = random.randint(0, SCREEN_WIDTH)
                new_enemy_y = random.randint(-100, -50)
                overlapping = False
                for existing_enemy in enemies:
                    if (new_enemy_x < existing_enemy.x + existing_enemy.size and
                            new_enemy_x + ENEMY_BASE_SIZE > existing_enemy.x and
                            new_enemy_y < existing_enemy.y + existing_enemy.size and
                            new_enemy_y + ENEMY_BASE_SIZE > existing_enemy.y):
                        overlapping = True
                        break
                if not overlapping:
                    break

            enemies.remove(enemy)  # Remove the old enemy
            enemies.append(Enemy(ENEMY_BASE_SIZE, ENEMY_BASE_SPEED))  # Add a new enemy
            total_score += SCORE_PER_WAVE
            wave_number += 1

    # Spawn more enemies on every 5th wave
    if wave_number % 5 == 0 and len(enemies) < wave_number // 5:
        enemies.append(Enemy(ENEMY_BASE_SIZE, ENEMY_BASE_SPEED))

    # Update ship cooldown
    ship.update_cooldown()

    # Schedule the next update
    window.after(16, update)


def draw():
    canvas.delete("all")

    # Draw stars
    for star in stars:
        sx, sy = star
        canvas.create_oval(sx, sy, sx + 2, sy + 2, fill="white")

    # Draw player's ship
    canvas.create_image((ship.x, ship.y), image=ship.image)

    # Draw enemies
    for enemy in enemies:
        canvas.create_image((enemy.x, enemy.y), image=enemy.image)

    # Display total score
    canvas.create_text(SCREEN_WIDTH - 50,
                       20,
                       text="Score: " + str(total_score),
                       fill="white",
                       anchor="e")
    canvas.create_text(SCREEN_WIDTH - 50,
                       40,
                       text="Wave: " + str(wave_number),
                       fill="white",
                       anchor="e")

    # Display elapsed time
    canvas.create_text(SCREEN_WIDTH - 50,
                       60,
                       text="Time: {:.1f}s".format(elapsed_time),
                       fill="white",
                       anchor="e")

    # Schedule the next draw
    window.after(16, draw)


def handle_key(event):
    if event.keysym in keys:
        keys[event.keysym] = True
    if event.keysym == 'space':
        ship.shoot()  # Shoot when spacebar is pressed


def handle_key_release(event):
    if event.keysym in keys:
        keys[event.keysym] = False


def game_over():
    global game_over_frame
    canvas.create_rectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, fill="red")
    game_over_frame = Frame(window)
    game_over_frame.place(relx=0.5, rely=0.5, anchor="center")
    game_over_label = Label(game_over_frame,
                            text="GAME OVER",
                            font=("Helvetica", 36),
                            fg="white",
                            bg="red")
    game_over_label.pack(pady=20)
    try_again_button = Button(game_over_frame,
                              text="Try Again",
                              command=restart_game)
    try_again_button.pack()


def restart_game():
    global game_over_frame, total_score, wave_number, ship, enemies, start_time
    game_over_frame.destroy()
    total_score = 0
    wave_number = 1
    ship = Ship()
    enemies = [Enemy(ENEMY_BASE_SIZE, ENEMY_BASE_SPEED)]
    start_time = time.time()  # Reset start time
    update()
    draw()


# Main Program
window = Tk()
window.title("Space Game")
canvas = Canvas(window, width=SCREEN_WIDTH, height=SCREEN_HEIGHT, bg="black")
canvas.pack()

keys = {"w": False, "a": False, "s": False, "d": False}

window.bind("<KeyPress>", handle_key)
window.bind("<KeyRelease>", handle_key_release)

ship = Ship()
enemies = [Enemy(ENEMY_BASE_SIZE, ENEMY_BASE_SPEED)]
start_time = time.time()  # Start time for elapsed time calculation

total_score = 0
wave_number = 1
elapsed_time = 0

# Generate stars
stars = []
for _ in range(100):
    sx = random.randint(0, SCREEN_WIDTH)
    sy = random.randint(0, SCREEN_HEIGHT)
    stars.append([sx, sy])

update()
draw()

window.mainloop()
